import sys
import itertools
from queue import PriorityQueue
import numpy as np

class Graph():

	def __init__(self, vertices=0):
		self.P = {}
		self.Pstar = {}
		self.vertices = vertices
		self.graph = np.full((self.vertices, self.vertices), np.inf)
		for x in range(self.vertices):
			self.graph[x][x] = 0

		for x in range(self.vertices):
			for y in range(self.vertices):
				self.P[(x,y)] = PriorityQueue()
				self.Pstar[(x,y)] = PriorityQueue()

		for x in range(self.vertices):
			info = {
					'path': ([x],[x]),
					'L': [],
					'Lstar': [],
					'R': [],
					'Rstar': []
					}
			self.P[(x,x)].put((0.0, info))
			self.Pstar[(x,x)].put((0.0, info))

	def _add_new_vertex(self):
			for x in range(self.vertices+1):
				self.P[(x,self.vertices)] = PriorityQueue()
				self.Pstar[(x,self.vertices)] = PriorityQueue()
			info = {
					'path': ([self.vertices],[self.vertices]),
					'L': [],
					'Lstar': [],
					'R': [],
					'Rstar': []
					}
			self.P[(self.vertices,self.vertices)].put((0.0, info))
			self.graph = np.concatenate([self.graph, np.full((1,self.vertices), np.inf)])
			self.vertices += 1
			self.graph = np.concatenate([self.graph, np.full((self.vertices, 1), np.inf)], axis=1)
			self.graph[self.vertices-1][self.vertices-1] = 0

	def distance(self, x, y):
		if x >= self.vertices or y >= self.vertices:
			return np.nan
		elif self.P[(x,y)].empty():
			return np.inf
		else:
			min_elem 	= self.P[(x,y)].get()
			min_weight 	= min_elem[0]
			self.P[(x,y)].put(min_elem)
			return min_weight

	def path(self, x, y):
		if x >= self.vertices or y >= self.vertices:
			print('No such vertex', file=sys.stderr)
			return
		elif self.P[(x,y)].empty():
			return []
		else:
			min_elem 	= self.P[(x,y)].get()
			min_path 	= min_elem[1]['path']
			self.P[(x,y)].put(min_elem)
			return (min_path[0] + [min_path[-1][-1]])

	def get(self, pi, elem, struct=None):
		if struct is None:
			struct = self.P
		start, end = pi[0][0], pi[-1][-1]
		items =list(filter(lambda x: x[1]['path']==pi, struct[(start, end)].queue))
		if items == []:
			return []
		return items[0][1][elem]

	def remove(self,pixy, star=False):
		x = pixy[0][0]
		y = pixy[-1][-1]

		struct = self.P[(x,y)]
		if star:
			struct = self.Pstar[(x,y)]

		info = list(filter(lambda item: item[1]['path']==pixy, struct.queue))[0]
		struct.queue.remove(info)

		only_one_vertex = len(pixy[0]) == 1 and pixy[0] == pixy[-1]

		if only_one_vertex:
			pass
		else:
			only_two_vertices = len(pixy[0]) == 1
			
			lxy = (pixy[0][:-1], pixy[0][1:])
			rxy = (pixy[1][:-1], pixy[1][1:])

			if only_two_vertices:
				lxy = ([pixy[0][0]], [pixy[0][0]])
				rxy = ([pixy[1][0]], [pixy[1][0]])

			remove_from = [(lxy,'R'),(rxy,'L')]
			if star:
				remove_from = [(lxy, 'Rstar'), (rxy, 'Lstar')]

			for path, array in remove_from:
				if pixy in self.get(path, array):
					self.get(path, array).remove(pixy)

	def update_graph(self, v, wdash):
		self.graph[:,v] = wdash[0]
		self.graph[v,:] = wdash[1]

	def cleanup(self, v):
		init_path  = ([v],[v])
		left_info  = self.get(init_path, 'L')
		right_info = self.get(init_path, 'R')
		Q = [(init_path, left_info, right_info)]
		over = [init_path]
		while Q != []:
			pi, left_info, right_info = Q.pop()
			for pixy in left_info:
				if pixy in over:
					continue
				over.insert(0, pixy)
				new_left_info  = self.get(pixy, 'L')
				new_right_info = self.get(pixy, 'R')
				Q.insert(0,(pixy,new_left_info,new_right_info))
				self.remove(pixy)
				if self.get(pixy, 'path', self.Pstar) != []:
					self.remove(pixy, True)

			for pixy in right_info:
				if pixy in over:
					continue
				over.insert(0, pixy)
				new_left_info  = self.get(pixy, 'L')
				new_right_info = self.get(pixy, 'R')
				Q.insert(0,(pixy,new_left_info,new_right_info))
				self.remove(pixy)
				if self.get(pixy, 'path', self.Pstar) != []:
					self.remove(pixy, True)

	def phase1(self, v, wdash):
		for u in range(self.vertices):
			if u == v:
				continue
			wuv = wdash[0][u]
			wvu = wdash[1][u]
			if wuv < np.inf:
				path 	= ([u],[v])
				weight 	= wuv
				info 	= {
							'path': path,
							'L': [],
							'Lstar': [],
							'R': [],
							'Rstar': []
						}
				self.P[(u,v)].put((wuv, info))
				self.get(([u],[u]), 'R').append(path)
				self.get(([v],[v]), 'L').append(path)
			if wvu < np.inf:
				path 	= ([v],[u])
				weight 	= wvu
				info 	= {
							'path': path,
							'L': [],
							'Lstar': [],
							'R': [],
							'Rstar': []
						}
				self.P[(v,u)].put((wvu, info))
				self.get(([v],[v]), 'R').append(path)
				self.get(([u],[u]), 'L').append(path)

	def phase2and3(self, v, wdash):
		#PHASE 2
		iter = itertools.count()
		h = PriorityQueue()
		already_marked = np.zeros((self.vertices, self.vertices))
		for x in range(self.vertices):
			for y in range(self.vertices):
				if not self.P[(x,y)].empty():
					min_elem = self.P[(x,y)].get()
					self.P[(x,y)].put(min_elem)
					h.put((min_elem[0], next(iter), min_elem[1]))

		#PHASE 3
		while not h.empty():
			extracted 		= h.get()
			(wxy,n,infoxy) 	= extracted
			to_put 			= (wxy, infoxy)
			pixy 			= infoxy['path']
			x, y 			= pixy[0][0], pixy[-1][-1]
			if already_marked[x][y] != 1:
				already_marked[x][y] = 1
				if self.Pstar[(x,y)].empty():
					self.Pstar[(x,y)].put(to_put)
					lxy = (pixy[0][:-1], pixy[0][1:])
					if len(lxy[0]) == 0:
						lxy = ([pixy[0][0]], [pixy[0][0]])
					self.get(lxy,'Rstar').append(pixy)
					rxy = (pixy[1][:-1], pixy[1][1:])
					if len(rxy[0]) == 0:
						rxy = ([pixy[-1][-1]], [pixy[-1][-1]])
					self.get(rxy,'Lstar').append(pixy)
					
					for pixdashb in self.get(lxy,'Lstar'):
						pixdashy = ( [pixdashb[0][0]] + pixy[0], [pixy[0][-1]] + pixy[-1] )
						weight   = self.graph[pixdashb[0][0]][pixy[0][0]] + wxy
						lxdashy  = pixdashb
						rxdashy  = pixy
						info 	 = {
									'path': pixdashy,
									'L': [],
									'Lstar': [],
									'R': [],
									'Rstar': []
								}
						self.P[(pixdashy[0][0],pixdashy[-1][-1])].put((weight,info))
						self.get(pixy, 'L').append(pixdashy)
						self.get(pixdashb, 'R').append(pixdashy)
						h.put((weight, next(iter), info))

					for piaydash in self.get(rxy, 'Rstar'):
						pixydash = ( [pixy[0][-1]] + pixy[-1], pixy[-1] + [piaydash[-1][-1]] )
						weight   = wxy + self.graph[pixy[-1][-1]][pixydash[-1][-1]]
						lxydash  = pixy
						rxydash  = piaydash
						info 	 = {
									'path': pixydash,
									'L': [],
									'Lstar': [],
									'R': [],
									'Rstar': []
								}
						self.P[(pixydash[0][0],pixydash[-1][-1])].put((weight,info))
						self.get(piaydash, 'L').append(pixydash)
						self.get(pixy, 'R').append(pixydash)
						h.put((weight, next(iter), info))


	def fixup(self, v, wdash):
		self.phase1(v, wdash)
		self.phase2and3(v, wdash)

	def update(self, v, wdash):
		self.update_graph(v, wdash)
		self.cleanup(v)
		self.fixup(v, wdash)

	def prin(self, which=None):
		if which:
			print(self.P[which].queue)
			return
		for (x,y) in self.P.items():
			if not y.empty():
				print(x, " : \n", y.queue)

	def Prin(self, which=None):
		if which:
			print(self.Pstar[which].queue)
			return
		for (x,y) in self.Pstar.items():
			if not y.empty():
				print(x, " : \n", y.queue)
